#ifndef CYGONCE_HAL_LEON3_SMP_H
#define CYGONCE_HAL_LEON3_SMP_H

//=============================================================================
//
//      hal_leon3_cmp.h
//
//      HAL smp defined
//
//=============================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//=============================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):   Gaisler Research (Konrad Eisele <eiselekd@web.de>)
// Contributors:
// Date:        2004-01-12
// Purpose:     Define Interrupt support
// Description:
// Usage:
//              #include <cyg/hal/hal_leon3.h>
//              ...
//
//
//####DESCRIPTIONEND####
//
//=============================================================================


#include <cyg/infra/cyg_type.h>         // Base types
#include <pkgconf/hal_sparc.h>
#include <cyg/hal/hal_leon3.h>

#include <cyg/kernel/gighook.h>

#ifdef CYGPKG_HAL_SMP_SUPPORT

//-----------------------------------------------------------------------------
// CPU numbering macros

externC cyg_uint32 cyg_hal_smp_cpu_running_count;

#define HAL_SMP_CPU_TYPE        cyg_uint32

#define HAL_SMP_CPU_MAX         CYGPKG_HAL_SMP_CPU_MAX

#define HAL_SMP_CPU_COUNT()     cyg_hal_smp_cpu_running_count

#define HAL_SMP_CPU_START_COUNT() HAL_SMP_CPU_MAX


externC cyg_uint32 cyg_hal_smp_cpu_count2idx(cyg_uint32 n);
externC cyg_uint32 cyg_hal_smp_cpu_idx2count(cyg_uint32 n);
#define HAL_SMP_CPU_COUNT2IDX(n)  cyg_hal_smp_cpu_count2idx(n)
#define HAL_SMP_CPU_IDX2COUNT(n)  cyg_hal_smp_cpu_idx2count(n)

externC cyg_uint32 sparc_leon3_get_cpuid(void);

#define HAL_SMP_CPU_THIS()                      \
({                                              \
    HAL_SMP_CPU_TYPE __id;                      \
    __asm__ __volatile__(                       \
         "rd     %%asr17,%0\n\t"                \
  : "=r" (__id) : );                      \
    ((__id >> 28) & 0xff);                      \
})

#define HAL_SMP_CPU_NONE        (CYGPKG_HAL_SMP_CPU_MAX+1)

//-----------------------------------------------------------------------------
// CPU startup

__externC cyg_uint32 cyg_hal_cpu_reset(cyg_uint32 cpu, cyg_uint32 arg2, cyg_uint32 arg3);

#define HAL_SMP_CPU_START( __cpu ) \
  (gighook[GIGHOOK_ECOS_HAL_SMP_CPU_START])((cyg_uint32)(__cpu),(cyg_uint32)(0),(cyg_uint32)(0));

//-----------------------------------------------------------------------------
// CPU message exchange

__externC CYG_WORD32 cyg_hal_cpu_message_isr( CYG_WORD32 vector, CYG_ADDRWORD data );
__externC CYG_WORD32 cyg_hal_cpu_message_dsr( CYG_WORD32 vector, CYG_ADDRWORD data );

__externC cyg_uint32 cyg_hal_cpu_message(cyg_uint32 cpu, cyg_uint32 msg, cyg_uint32 wait);

#define HAL_SMP_MESSAGE_TYPE            0xF0000000
#define HAL_SMP_MESSAGE_ARG             (~HAL_SMP_MESSAGE_TYPE)

#define HAL_SMP_MESSAGE_RESCHEDULE      0x10000000
#define HAL_SMP_MESSAGE_MASK            0x20000000
#define HAL_SMP_MESSAGE_UNMASK          0x30000000
#define HAL_SMP_MESSAGE_REVECTOR        0x40000000
#define HAL_SMP_MESSAGE_TIMESLICE       0x50000000

#define HAL_SMP_CPU_RESCHEDULE_INTERRUPT( __cpu, __wait ) \
  (gighook[GIGHOOK_ECOS_HAL_SMP_CPU_MESSAGE])((cyg_uint32)(__cpu),(cyg_uint32)(HAL_SMP_MESSAGE_RESCHEDULE),(cyg_uint32)(__wait));

#define HAL_SMP_CPU_TIMESLICE_INTERRUPT( __cpu, __wait ) \
  (gighook[GIGHOOK_ECOS_HAL_SMP_CPU_MESSAGE])((cyg_uint32)(__cpu),(cyg_uint32)(HAL_SMP_MESSAGE_TIMESLICE),(cyg_uint32)(__wait));

//-----------------------------------------------------------------------------
// Test-and-set support
// These macros provide test-and-set support for the least significant bit
// in a word.

#define HAL_TAS_TYPE    volatile CYG_WORD32


#ifndef _NO_LDD_

#define HAL_TAS_SET( _tas_, _oldb_ )                    \
CYG_MACRO_START                                         \
{                                                       \
    register CYG_WORD32 __old = 1;                      \
    __asm__ volatile ("swapa [%2] %3, %0" :             \
            "=r" (__old) :            \
            "0" (__old),              \
            "r" (&_tas_),             \
            "i" (1));                 \
    _oldb_ = ( __old & 1 ) != 0;                        \
}                                                       \
CYG_MACRO_END

#define HAL_TAS_CLEAR( _tas_, _oldb_ )                  \
CYG_MACRO_START                                         \
{                                                       \
    register CYG_WORD32 __old = 0;                      \
    __asm__ volatile ("swapa [%2] %3, %0" :             \
            "=r" (__old) :            \
            "0" (__old),              \
            "r" (&_tas_),             \
            "i" (1));                 \
    _oldb_ = ( __old & 1 ) != 0;                        \
}                                                       \
CYG_MACRO_END

#else /* _NO_LDD_ */

/* HAL_TAS_SET workaround for swapa bug on lewis */

#define CACHE_LINE_ALIGNED  __attribute__ ((section (".cache_aligned")))

externC CYG_WORD32 __hal_tas_set( volatile CYG_WORD32 * _tas_,
                                    CYG_WORD32  __oldb) CACHE_LINE_ALIGNED;


#define HAL_TAS_SET( _tas_, _oldb_ )                  \
CYG_MACRO_START                                       \
{                                                     \
    register CYG_WORD32 __old = 1;                    \
    __old=__hal_tas_set(&_tas_ , __old);              \
    _oldb_ = ( __old & 1 ) != 0;                      \
}                                                     \
CYG_MACRO_END

#define HAL_TAS_CLEAR( _tas_, _oldb_ )                \
CYG_MACRO_START                                       \
{                                                     \
    register CYG_WORD32 __old = 0;                    \
    __old=__hal_tas_set(&_tas_ , __old);              \
    _oldb_ = ( __old & 1 ) != 0;                      \
}                                                     \
CYG_MACRO_END


#endif

//-----------------------------------------------------------------------------
// Spinlock support.
// Built on top of test-and-set code.

#define HAL_SPINLOCK_TYPE       volatile CYG_WORD32

#define HAL_SPINLOCK_INIT_CLEAR 0

#define HAL_SPINLOCK_INIT_SET   1

#define HAL_SPINLOCK_SPIN( _lock_ )             \
CYG_MACRO_START                                 \
{                                               \
    cyg_bool _val_;                             \
    do                                          \
    {                                           \
        HAL_TAS_SET( _lock_, _val_ );           \
    } while( _val_ );                           \
}                                               \
CYG_MACRO_END

#define HAL_SPINLOCK_CLEAR( _lock_ )            \
CYG_MACRO_START                                 \
{                                               \
    cyg_bool _val_;                             \
    HAL_TAS_CLEAR( _lock_ , _val_ );            \
}                                               \
CYG_MACRO_END

#define HAL_SPINLOCK_TRY( _lock_, _val_ )       \
    HAL_TAS_SET( _lock_, _val_ );               \
    (_val_) = (((_val_) & 1) == 0)

#define HAL_SPINLOCK_TEST( _lock_, _val_ )      \
    (_val_) = (((_lock_) & 1) != 0)

//-----------------------------------------------------------------------------
// Diagnostic output serialization

/*__externC HAL_SPINLOCK_TYPE cyg_hal_smp_diag_lock;

#define CYG_HAL_DIAG_LOCK_DATA_DEFN \
        HAL_SPINLOCK_TYPE cyg_hal_smp_diag_lock = HAL_SPINLOCK_INIT_CLEAR

#define CYG_HAL_DIAG_LOCK() HAL_SPINLOCK_SPIN( cyg_hal_smp_diag_lock )

#define CYG_HAL_DIAG_UNLOCK() HAL_SPINLOCK_CLEAR( cyg_hal_smp_diag_lock )

#define leon3smp_diag_printf(fmt,arg...)  \
    CYG_HAL_DIAG_LOCK();    \
    diag_printf(fmt,##arg); \
    CYG_HAL_DIAG_UNLOCK();*/

#define leon3smp_diag_printf(fmt,arg...)

//-----------------------------------------------------------------------------
// Some extra definitions

__externC HAL_SPINLOCK_TYPE cyg_hal_ioapic_lock;

//-----------------------------------------------------------------------------

//--------------------------------------------------------------------------
// irqmp interrupt configuration
// Additional SMP interrupt configuration support.

externC cyg_uint32 leon3_smpirq_init(cyg_uint32 arg1, cyg_uint32 arg2, cyg_uint32 arg3);
externC cyg_uint32 leon3_interrupt_set_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 );
externC cyg_uint32 leon3_interrupt_get_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 );
externC cyg_uint32 leon3_smpirq_raise(cyg_uint32 cpu, cyg_uint32 arg2, cyg_uint32 arg3);

externC cyg_uint32 new_leon3_interrupt_set_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 ) __attribute__ ((section (".forceram")));

externC void leon3_smpirq_lock(void);
externC void leon3_smpirq_unlock(void);

//#define HAL_INTERRUPT_SMP_INIT( ) leon3_smpirq_init( );
//#define HAL_INTERRUPT_SET_CPU( _vector_, _cpu_ )  leon3_interrupt_set_cpu( _vector_, _cpu_ );
//#define HAL_INTERRUPT_GET_CPU( _vector_, _cpu_ )  leon3_interrupt_get_cpu( _vector_, &(_cpu_) );
//#define HAL_INTERRUPT_SMP_RAISE( _cpu_ ) leon3_smpirq_raise( _cpu_ );

#define HAL_INTERRUPT_SMP_INIT( ) \
  (gighook[GIGHOOK_ECOS_HAL_INTERRUPT_SMP_INIT])((cyg_uint32)(0),(cyg_uint32)(0),(cyg_uint32)(0))

#define HAL_INTERRUPT_SET_CPU( _vector_, _cpu_ ) \
  (gighook[GIGHOOK_ECOS_HAL_INTERRUPT_SET_CPU])((cyg_uint32)(_vector_),(cyg_uint32)(_cpu_),(cyg_uint32)(0))

#define HAL_INTERRUPT_GET_CPU( _vector_, _cpu_ ) \
  (gighook[GIGHOOK_ECOS_HAL_INTERRUPT_GET_CPU])((cyg_uint32)(_vector_),(cyg_uint32)(_cpu_),(cyg_uint32)(0))

#define HAL_INTERRUPT_SMP_RAISE( _cpu_ ) \
  (gighook[GIGHOOK_ECOS_HAL_INTERRUPT_SMP_RAISE])((cyg_uint32)(_cpu_),(cyg_uint32)(0),(cyg_uint32)(0))

#endif // CYGPKG_HAL_SMP_SUPPORT

//-----------------------------------------------------------------------------
#endif // ifndef CYGONCE_HAL_LEON3_SMP_H
// End of hal_leon3_smp.h
