//===========================================================================
//
//      hal_amba.c
//
//      SPARC Architecture LEON3-specific private variables
//
//===========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Gaisler Research (Konrad Eisele<eiselekd@web.de>)
// Contributors:
// Date:         2004
// Purpose:      private amba for LEON3 processor.
//
//####DESCRIPTIONEND####
//
//===========================================================================

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <cyg/hal/hal_leon3.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/hal_intr.h>

//#define DEBUG_CONFIG

/* Structure containing address to devices found on the Amba Plug&Play bus */
amba_confarea_type amba_conf;

#ifdef _GRSIM_

/* Pointers to Interrupt Controller configuration registers */
volatile LEON3_IrqCtrl_Regs_Map *LEON3_IrqCtrl_Regs = 0;
volatile LEON3_GpTimer_Regs_Map *LEON3_GpTimer_Regs = 0;
volatile LEON3_APBUART_Regs_Map *LEON3_APBUART_Regs = 0;
unsigned long LEON3_GpTimer_Irq = 0;

#else /* _GRSIM_ */

/* Pointers to Interrupt Controller configuration registers */
volatile LEON3_IrqCtrl_Regs_Map_Read *LEON3_IrqCtrl_Regs_Read = 0;
volatile LEON3_IrqCtrl_Regs_Map_Write *LEON3_IrqCtrl_Regs_Write = 0;
volatile LEON3_GpTimer_Regs_Map *LEON3_GpTimer_2_Regs = 0;
volatile LEON3_APBUART_Regs_Map *LEON3_APBUART_Regs = 0;
unsigned long LEON3_GpTimer_2_Irq = 0;

#endif /* _GRSIM_ */

//static void vendor_dev_string(unsigned long conf, char *vendorbuf,char *devbuf) {
//  int vendor = amba_vendor(conf); int dev = amba_device(conf);
//  char *devstr; char *vendorstr;
//  sprintf(vendorbuf, "Unknown vendor %2x",vendor);
//  sprintf(devbuf, "Unknown device %2x",dev);
//  vendorstr = vendor_id2str(vendor);
//  if (vendorstr) {
//    sprintf(vendorbuf, "%s",vendorstr);
//  }
//  devstr = device_id2str(vendor,dev);
//  if (devstr) {
//    sprintf(devbuf, "%s",devstr);
//  }
//}

void amba_prinf_config(void)
{
  unsigned int conf;
  //char devbuf[128]; char vendorbuf[128];
  int i = 0; int j = 0;
  unsigned int addr; unsigned int m;
  //diag_printf("             Vendors         Slaves\n");
  //diag_printf("Ahb masters:\n");
  i = 0;
  while (i < amba_conf.ahbmst.devnr)
  {
    conf = amba_get_confword(amba_conf.ahbmst, i, 0);
    //vendor_dev_string(conf,vendorbuf,devbuf);
    //diag_printf("%2i(%2x:%3x|%2i): %16s %16s \n", i, amba_vendor(conf), amba_device(conf), amba_irq(conf), vendorbuf, devbuf);
    for (j = 0;j < 4;j++) {
      m = amba_ahb_get_membar(amba_conf.ahbmst,i,j);
      if (m) {
        addr = amba_membar_start(m);
        //diag_printf(" +%i: 0x%x \n", j, addr);
      }
    }
    i++;
  }
  //diag_printf("Ahb slaves:\n");
  i = 0;
  while (i < amba_conf.ahbslv.devnr)
  {
    conf = amba_get_confword(amba_conf.ahbslv, i, 0);
    //vendor_dev_string(conf,vendorbuf,devbuf);
    //diag_printf("%2i(%2x:%3x|%2i): %16s %16s \n", i, amba_vendor(conf), amba_device(conf), amba_irq(conf), vendorbuf, devbuf);
    for (j = 0;j < 4;j++) {
      m = amba_ahb_get_membar(amba_conf.ahbslv,i,j);
      if (m) {
        addr = amba_membar_start(m);
  if (amba_membar_type(m) == AMBA_TYPE_AHBIO) {
    addr = AMBA_TYPE_AHBIO_ADDR(addr);
  } else if (amba_membar_type(m) == AMBA_TYPE_APBIO) {
    //diag_printf("Warning: apbio membar\n");
  }
  //diag_printf(" +%i: 0x%x (raw:0x%x)\n", j, addr, m);
      }
    }
    i++;
  }
  //diag_printf("Apb slaves:\n");
  i = 0;
  while (i < amba_conf.apbslv.devnr)
  {

    conf = amba_get_confword(amba_conf.apbslv, i, 0);
    //vendor_dev_string(conf,vendorbuf,devbuf);
    //diag_printf("%2i(%2x:%3x|%2i): %16s %16s \n", i, amba_vendor(conf), amba_device(conf), amba_irq(conf), vendorbuf, devbuf);

    m = amba_apb_get_membar(amba_conf.apbslv, i);
    addr = amba_iobar_start(amba_conf.apbmst, m);
    //diag_printf(" +%2i: 0x%x (raw:0x%x) \n", 0, addr, m);

    i++;

  }

}

#define amba_insert_device(tab, address) \
{ \
  if (LEON3_BYPASS_LOAD_PA(address)) \
  { \
    (tab)->addr[(tab)->devnr] = (address); \
    (tab)->devnr ++; \
  } \
} while(0)

/*
 *  Used to scan system bus. Probes for AHB masters, AHB slaves and
 *  APB slaves. Addresses to configuration areas of the AHB masters,
 *  AHB slaves, APB slaves and APB master are storeds in
 *  amba_ahb_masters, amba_ahb_slaves and amba.
 */

void amba_init(void)
{
  unsigned int *cfg_area;  /* address to configuration area */
  unsigned int mbar, conf;
  int i, j;

#ifdef DEBUG_CONFIG
  diag_printf("Reading AMBA Plug&Play configuration area\n");
#endif

  memset(&amba_conf,0,sizeof(amba_conf));
  //amba_conf.ahbmst.devnr = 0; amba_conf.ahbslv.devnr = 0; amba_conf.apbslv.devnr = 0;

  cfg_area = (unsigned int *) (LEON3_IO_AREA | LEON3_CONF_AREA);

  for (i = 0; i < LEON3_AHB_MASTERS; i++)
  {
    amba_insert_device(&amba_conf.ahbmst, cfg_area);
    cfg_area += LEON3_AHB_CONF_WORDS;
  }

  cfg_area = (unsigned int *) (LEON3_IO_AREA | LEON3_CONF_AREA | LEON3_AHB_SLAVE_CONF_AREA);
  for (i = 0; i < LEON3_AHB_SLAVES; i++)
  {
    amba_insert_device(&amba_conf.ahbslv, cfg_area);
    cfg_area += LEON3_AHB_CONF_WORDS;
  }

  for (i = 0; i < amba_conf.ahbslv.devnr; i ++)
  {
    conf = amba_get_confword(amba_conf.ahbslv, i, 0);
    mbar = amba_ahb_get_membar(amba_conf.ahbslv, i, 0);
    if ((amba_vendor(conf) == VENDOR_GAISLER) && (amba_device(conf) == GAISLER_APBMST))
    {
      amba_conf.apbmst = amba_membar_start(mbar);
      cfg_area = (unsigned int *) (amba_conf.apbmst | LEON3_CONF_AREA);

      //diag_printf("Found apbmst, cfg: 0x%x\n",(unsigned int)cfg_area);

      for (j = amba_conf.apbslv.devnr; j < LEON3_APB_SLAVES; j++)
      {
  amba_insert_device(&amba_conf.apbslv, cfg_area);
  cfg_area += LEON3_APB_CONF_WORDS;
      }
      break;
    }
  }

  /* Call to specific actions after device detection. */
  (gighook[GIGHOOK_ECOS_AMBA_INIT_POST])((cyg_uint32)(0),(cyg_uint32)(0),(cyg_uint32)(0));

}


cyg_uint32 amba_init_gigle(cyg_uint32 arg1, cyg_uint32 arg2, cyg_uint32 arg3)
{

#ifdef _GRSIM_

  /* Find LEON3 Interrupt controler */
  LEON3_IrqCtrl_Regs = (volatile LEON3_IrqCtrl_Regs_Map *) amba_find_apbslv_addr(VENDOR_GAISLER, GAISLER_IRQMP, 0);
  LEON3_GpTimer_Regs = (volatile LEON3_GpTimer_Regs_Map *) amba_find_apbslv_addr(VENDOR_GAISLER, GAISLER_GPTIMER, &LEON3_GpTimer_Irq);
  LEON3_APBUART_Regs = (volatile LEON3_APBUART_Regs_Map *) 0x00400100;
  if (LEON3_IrqCtrl_Regs) {
    //LEON3_BYPASS_STORE_PA(&(LEON3_IrqCtrl_Regs ->mask[0]),0);
  }

#else /* _GRSIM_ */

  /* Find LEON3 Interrupt controler */
  LEON3_IrqCtrl_Regs_Read = (volatile LEON3_IrqCtrl_Regs_Map_Read *) amba_find_apbslv_addr(VENDOR_GIGLE, GIGLE_IRQMP, 0);
  LEON3_IrqCtrl_Regs_Write = (volatile LEON3_IrqCtrl_Regs_Map_Write *) LEON3_IrqCtrl_Regs_Read;
  /* Fix address of GpTimer */
  LEON3_GpTimer_2_Regs = (volatile LEON3_GpTimer_Regs_Map *) 0x00400D00;
  LEON3_GpTimer_2_Irq = CYGNUM_HAL_INTERRUPT(5);
//  if (LEON3_IrqCtrl_Regs) {
//    //LEON3_BYPASS_STORE_PA(&(LEON3_IrqCtrl_Regs ->mask[0]),0);
//  }
  LEON3_GpTimer_2_Regs->scalar = ((AMBA_FREQ_KHZ+500)/1000)-1;
  LEON3_GpTimer_2_Regs->scalar_reload = ((AMBA_FREQ_KHZ+500)/1000)-1;

#endif /* _GRSIM_ */

  return 0;
}

unsigned long amba_find_apbslv_addr(unsigned long vendor, unsigned long device, unsigned long *irq) {
  unsigned int i,conf,iobar;
  for (i = 0; i < amba_conf.apbslv.devnr; i++) {
    conf = amba_get_confword(amba_conf.apbslv, i, 0);
    if ((amba_vendor(conf) == vendor) && (amba_device(conf) == device))
    {
      if (irq) {
        *irq = amba_irq(conf);
      }
      iobar = amba_apb_get_membar(amba_conf.apbslv, i);
      return amba_iobar_start(amba_conf.apbmst, iobar);
    }
  }
  return 0;
}
