//===========================================================================
//
//      hal_irq.c
//
//      SPARC Architecture LEON3-specific private variables
//
//===========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Gaisler Research (Konrad Eisele<eiselekd@web.de>)
// Contributors:
// Date:         2004
// Purpose:      private amba for LEON3 processor.
//
//####DESCRIPTIONEND####
//
//===========================================================================

#ifdef CYGPKG_KERNEL
#include <pkgconf/kernel.h>
#endif

#include <cyg/infra/cyg_type.h>         // Base types
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros
#include <cyg/infra/diag.h>
#include <cyg/hal/drv_api.h>

#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_leon3.h>
#include <cyg/hal/hal_smp.h>

#include <cyg/kernel/gighook.h>

#ifdef _GRSIM_

#define LEON_HARD_INT(x) (1 << (x))
externC volatile LEON3_IrqCtrl_Regs_Map *LEON3_IrqCtrl_Regs;
#define LEON_IMASK(i) ((&LEON3_IrqCtrl_Regs ->mask[i]))
#define LEON_IFORCE(i) ((&LEON3_IrqCtrl_Regs ->force[i]))

#define get_irqmask(irq) LEON_HARD_INT(irq)

/*static inline unsigned long get_irqmask(unsigned int irq)
{
  unsigned long mask;
  if (!irq || irq > 0xf) {
    //diag_printf("leon_get_irqmask: false irq number\n");
    mask = 0;
  } else {
    mask = LEON_HARD_INT(irq);
  }
  return mask;
}*/

#ifndef CYGPKG_HAL_SMP_SUPPORT

void leon3_disable_irq(unsigned int irq_nr)
{
  unsigned long mask, v;
  mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;
  HAL_READ_UINT32(LEON_IMASK(0),v);
  HAL_WRITE_UINT32(LEON_IMASK(0), (v & ~(mask)));

}

void leon3_enable_irq(unsigned int irq_nr)
{
  unsigned long mask,v;
  mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;
  HAL_READ_UINT32(LEON_IMASK(0),v);
  HAL_WRITE_UINT32(LEON_IMASK(0), (v | (mask)));
}

#else

unsigned long irq_mask;
unsigned long cpu_route_mask[CYGPKG_HAL_SMP_CPU_MAX];
HAL_SPINLOCK_TYPE cyg_hal_smp_smpirq_lock = HAL_SPINLOCK_INIT_CLEAR;

/* route all irq to 0 */
cyg_uint32 leon3_smpirq_init(cyg_uint32 arg1, cyg_uint32 arg2, cyg_uint32 arg3) {

  cpu_route_mask[0] = -1;

  return 0;
}

cyg_uint32 leon3_disable_irq(cyg_uint32 irq_nr, cyg_uint32 arg2, cyg_uint32 arg3)
{
  unsigned long mask, cpu;

  mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;
  irq_mask &= ~mask;
  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );
  for (cpu = 0; cpu < HAL_SMP_CPU_COUNT(); cpu++) {
    mask = cpu_route_mask[cpu] & irq_mask;
    HAL_WRITE_UINT32(LEON_IMASK(cpu), mask);
  }
  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  return 0;
}

cyg_uint32 leon3_enable_irq(cyg_uint32 irq_nr, cyg_uint32 arg2, cyg_uint32 arg3)
{
  unsigned long mask, cpu, v = 0;
  mask = get_irqmask(irq_nr) & LEON_IRQMASK_R;

  for (cpu = 0; cpu < HAL_SMP_CPU_MAX; cpu++) {
    v |= cpu_route_mask[cpu] & mask;
  }
  if (!v) {
    cpu_route_mask[0] |= mask;
  }

  irq_mask |= mask;
  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );
  for (cpu = 0; cpu < HAL_SMP_CPU_MAX; cpu++) {
    mask = (cpu_route_mask[cpu] | 0x04000) & irq_mask;
    HAL_WRITE_UINT32(LEON_IMASK(cpu), mask);
  }
  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  return 0;
}

cyg_uint32 leon3_interrupt_set_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 ) {
  unsigned long mask, i;

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );
  mask = get_irqmask(vector) & LEON_IRQMASK_R;
  for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) {
    if ( i == cpu) {
      cpu_route_mask[i] |= mask;
    } else {
      cpu_route_mask[i] &= ~mask;
    }
    HAL_WRITE_UINT32(LEON_IMASK(i), cpu_route_mask[i] & irq_mask);
  }
  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  return 0;
}

cyg_uint32 leon3_interrupt_get_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 ) {
  unsigned long mask, i;

  mask = get_irqmask(vector) & LEON_IRQMASK_R;
  *((cyg_uint32 *)cpu) = 0;
  for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) {
    if (cpu_route_mask[i] & mask) {
      *((cyg_uint32 *)cpu) = i;
    }
  }

  return 0;
}

cyg_uint32 leon3_smpirq_raise(cyg_uint32 cpu, cyg_uint32 arg2, cyg_uint32 arg3) {
    unsigned long mask = get_irqmask(CYGNUM_HAL_SMP_CPU_INTERRUPT_VECTOR(cpu)) & LEON_IRQMASK_R;

    CYG_INTERRUPT_STATE istate = 0;

    HAL_DISABLE_INTERRUPTS( istate );

    HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );

    HAL_WRITE_UINT32(LEON_IFORCE(cpu), (mask));

    HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

    HAL_RESTORE_INTERRUPTS( istate );

    return 0;
}

#endif

#else /* _GRSIM_ */

// Get mask of a given interrupt
#define LEON_BANK_INT(x) \
          (1 << (((x) - CYGNUM_HAL_ISR_MIN) % CYGNUM_HAL_INTERRUPS_PER_BANK))

// Get bank of a given interrupt - number of banks must be power of two
#define CYGNUM_HAL_BANK_GET(_n_) \
          ((((_n_) - CYGNUM_HAL_ISR_MIN) / CYGNUM_HAL_INTERRUPS_PER_BANK) + 1)

externC volatile LEON3_IrqCtrl_Regs_Map_Write *LEON3_IrqCtrl_Regs_Write;
externC volatile LEON3_IrqCtrl_Regs_Map_Read *LEON3_IrqCtrl_Regs_Read;
#define LEON_ENABLE_SET(cpu,bank)   \
   ((&LEON3_IrqCtrl_Regs_Write ->set_enable[cpu*CYGNUM_HAL_BANK_COUNT+bank]))
#define LEON_ENABLE_GET(cpu,bank)   \
   ((&LEON3_IrqCtrl_Regs_Read ->enable[cpu*CYGNUM_HAL_BANK_COUNT+bank]))
#define LEON_PENDING_GET(cpu,bank)   \
   ((&LEON3_IrqCtrl_Regs_Read ->pending[cpu*CYGNUM_HAL_BANK_COUNT+bank]))

#define LEON_ENABLE_CLEAR(cpu,bank) \
   ((&LEON3_IrqCtrl_Regs_Write ->clear_enable[cpu*CYGNUM_HAL_BANK_COUNT+bank]))
#define LEON_RAW_FORCE(bank)    \
   ((&LEON3_IrqCtrl_Regs_Write ->force_raw[bank]))
#define LEON_RAW_RELEASE(bank)    \
   ((&LEON3_IrqCtrl_Regs_Write ->release_raw[bank]))

externC volatile LEON3_IrqCtrl_Regs_Map_Read *LEON3_IrqCtrl_Regs_Read;
#define LEON_PENDING(cpu,bank)   \
   ((&LEON3_IrqCtrl_Regs_Read ->pending[cpu*CYGNUM_HAL_BANK_COUNT+bank]))

#define get_irqmask(irq) LEON_BANK_INT(irq)


static unsigned long irq_mask[CYGNUM_HAL_BANK_COUNT];
static unsigned long cpu_route_mask[CYGPKG_HAL_SMP_CPU_MAX][CYGNUM_HAL_BANK_COUNT];
static HAL_SPINLOCK_TYPE cyg_hal_smp_smpirq_lock = HAL_SPINLOCK_INIT_CLEAR;

cyg_uint32 leon3_smpirq_init(cyg_uint32 arg1, cyg_uint32 arg2, cyg_uint32 arg3) {
  int i,j;
  unsigned int cpu_irq_nr;

  /* Initialize interrupt tables:
   * 1) hardware interrupts
   **/
  for (j=0; j<CYGNUM_HAL_EXCEPTION_MIN; j++)
  {
    hal_interrupt_handlers[j] = (CYG_ADDRESS)hal_default_isr;
    hal_interrupt_data[j] = 0x11da1a00 + j;
    hal_interrupt_objects[j] = (CYG_ADDRESS)NULL;
  }

  /* 2) exception vectors
   **/
  for (j=CYGNUM_HAL_EXCEPTION_MIN; j<=CYGNUM_HAL_EXCEPTION_MAX; j++)
  {
    hal_interrupt_handlers[j] = (CYG_ADDRESS)cyg_hal_exception_handler;
    hal_interrupt_data[j] = 0xeeda1a00 - CYGNUM_HAL_EXCEPTION_MIN + j;
    hal_interrupt_objects[j] = (CYG_ADDRESS)NULL;
  }

  /* 3) source interrupts
   **/
  for (j=CYGNUM_HAL_ISR_MIN; j<=CYGNUM_HAL_ISR_MAX; j++)
  {
    hal_interrupt_handlers[j] = (CYG_ADDRESS)hal_default_source_isr;
    hal_interrupt_data[j] = 0x22da1a00 - CYGNUM_HAL_ISR_MIN + j;
    hal_interrupt_objects[j] = (CYG_ADDRESS)NULL;
  }

  /* No interrupts enabled. */
  for (j=0; j<CYGNUM_HAL_BANK_COUNT; j++)
  {
    irq_mask[j] = 0;
  }

  /* All interrupts are enabled in CPU 0. */
  for (j=0; j<CYGNUM_HAL_BANK_COUNT; j++)
  {
    cpu_route_mask[0][j] = 0xFFFFFFFF;
  }

  /* No interrupts enabled in the other CPU's...*/
  for (i=1; i<CYGPKG_HAL_SMP_CPU_MAX; i++)
  {
    for (j=0; j<CYGNUM_HAL_BANK_COUNT; j++)
    {
      cpu_route_mask[i][j] = 0;
    }
  }

  /*  ...except for CPU message. */
  for (i=0; i<CYGPKG_HAL_SMP_CPU_MAX; i++)
  {
    cpu_irq_nr = CYGNUM_HAL_SMP_CPU_INTERRUPT_VECTOR(i);
    cpu_route_mask[i][CYGNUM_HAL_BANK_GET(cpu_irq_nr)] |= get_irqmask(cpu_irq_nr);
    for (j=0; j<CYGPKG_HAL_SMP_CPU_MAX; j++)
    {
      if (i != j)
      {
        cpu_route_mask[j][CYGNUM_HAL_BANK_GET(cpu_irq_nr)] &= ~(get_irqmask(cpu_irq_nr));
      }
    }
  }

return 0;

}

cyg_uint32 leon3_disable_irq(cyg_uint32 irq_nr, cyg_uint32 arg2, cyg_uint32 arg3)
{
  unsigned long bank, mask, cpu;

  CYG_INTERRUPT_STATE istate = 0;

  HAL_DISABLE_INTERRUPTS( istate );

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );

  bank = CYGNUM_HAL_BANK_GET(irq_nr);
  mask = get_irqmask(irq_nr);

  irq_mask[bank] &= ~mask;
  for (cpu = 0; cpu < CYGPKG_HAL_SMP_CPU_MAX; cpu++) {
    /* The interrupt will only be cleared in the routed CPU, the other ones
     * should already be disabled */
    HAL_WRITE_UINT32(LEON_ENABLE_CLEAR(cpu,bank),
                     mask & cpu_route_mask[cpu][bank]);
  }

  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  HAL_RESTORE_INTERRUPTS( istate );

  return 0;
}

cyg_uint32 leon3_enable_irq(cyg_uint32 irq_nr, cyg_uint32 arg2, cyg_uint32 arg3)
{
  unsigned long bank, mask, cpu;

  CYG_INTERRUPT_STATE istate = 0;

  HAL_DISABLE_INTERRUPTS( istate );

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );

  bank = CYGNUM_HAL_BANK_GET(irq_nr);
  mask = get_irqmask(irq_nr);

// These following lines are necessary due to the current start-up sequence.
// The cpu_route_mask is deleted during the initialization, so the first enable_irq
// call will reset it to the default values.
// Should be fixed if the start-up sequence is revised.

  {
    unsigned long v = 0;
    for (cpu = 0; cpu < HAL_SMP_CPU_MAX; cpu++) {
      v |= cpu_route_mask[cpu][bank] & mask;
    }
    if (!v) {
      cpu_route_mask[0][bank] |= mask;
    }
  }

  irq_mask[bank] |= mask;
  for (cpu = 0; cpu < CYGPKG_HAL_SMP_CPU_MAX; cpu++) {
    /* The interrupt will only be set in the routed CPU, the other ones
     * should already be disabled */
    HAL_WRITE_UINT32(LEON_ENABLE_SET(cpu,bank),
                     mask & cpu_route_mask[cpu][bank]);
  }

  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  HAL_RESTORE_INTERRUPTS( istate );

  return 0;
}

cyg_uint32 leon3_interrupt_set_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 )
{
  unsigned long bank, mask, i;

  CYG_INTERRUPT_STATE istate = 0;

  HAL_DISABLE_INTERRUPTS( istate );

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );

  bank = CYGNUM_HAL_BANK_GET(vector);
  mask = get_irqmask(vector);

  {
    for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) {
      if ( i == cpu) {
        cpu_route_mask[i][bank] |= mask;
        HAL_WRITE_UINT32(LEON_ENABLE_SET(cpu,bank), mask & irq_mask[bank]);
      } else {
        cpu_route_mask[i][bank] &= ~mask;
        HAL_WRITE_UINT32(LEON_ENABLE_CLEAR(cpu,bank), mask);
      }
    }
  }

  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  HAL_RESTORE_INTERRUPTS( istate );

  return 0;
}

cyg_uint32 new_leon3_interrupt_set_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 )
{
  unsigned long bank, mask, i;

  CYG_INTERRUPT_STATE istate = 0;

  HAL_DISABLE_INTERRUPTS( istate );

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );

  bank = CYGNUM_HAL_BANK_GET(vector);
  mask = get_irqmask(vector);

  {
    for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) {
      if ( i == cpu) {
        cpu_route_mask[i][bank] |= mask;
        HAL_WRITE_UINT32(LEON_ENABLE_SET(i,bank), mask & irq_mask[bank]);
      } else {
        cpu_route_mask[i][bank] &= ~mask;
        HAL_WRITE_UINT32(LEON_ENABLE_CLEAR(i,bank), mask);
      }
    }
  }

  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  HAL_RESTORE_INTERRUPTS( istate );

  return 0;
}

cyg_uint32 leon3_interrupt_get_cpu( cyg_uint32 vector, cyg_uint32 cpu, cyg_uint32 arg3 )
{
  unsigned long bank, mask, i;

  bank = CYGNUM_HAL_BANK_GET(vector);
  mask = get_irqmask(vector);

  *((cyg_uint32 *)cpu) = 0;
  for (i = 0; i < HAL_SMP_CPU_COUNT(); i++) {
    if (cpu_route_mask[i][bank] & mask) {
      *((cyg_uint32 *)cpu) = i;
    }
  }
  return 0;
}
int cpu_int[2] = {0,0};
cyg_uint32 leon3_smpirq_raise(cyg_uint32 cpu, cyg_uint32 arg2, cyg_uint32 arg3)
{
  unsigned long bank, mask;

  cpu_int[cpu]++;
  bank = CYGNUM_HAL_BANK_GET(CYGNUM_HAL_SMP_CPU_INTERRUPT_VECTOR(cpu));
  mask = get_irqmask(CYGNUM_HAL_SMP_CPU_INTERRUPT_VECTOR(cpu));

  CYG_INTERRUPT_STATE istate = 0;

  HAL_DISABLE_INTERRUPTS( istate );

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );

  // Force interrupt
  HAL_WRITE_UINT32(LEON_RAW_FORCE(bank), mask);

  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  HAL_RESTORE_INTERRUPTS( istate );

  return 0;
}

cyg_uint32 leon3_acknowledge_irq(cyg_uint32 irq_nr, cyg_uint32 arg2, cyg_uint32 arg3)
{
  unsigned long bank, mask;

  CYG_INTERRUPT_STATE istate = 0;

  HAL_DISABLE_INTERRUPTS( istate );

  HAL_SPINLOCK_SPIN( cyg_hal_smp_smpirq_lock );

  bank = CYGNUM_HAL_BANK_GET(irq_nr);
  mask = get_irqmask(irq_nr);

  HAL_WRITE_UINT32(LEON_RAW_RELEASE(bank), mask);

  HAL_SPINLOCK_CLEAR( cyg_hal_smp_smpirq_lock );

  HAL_RESTORE_INTERRUPTS( istate );

  return 0;
}

// New default ISR: decode interrupt source and call corresponding ISR
//                  (& post DSR if necessary)
volatile CYG_ADDRWORD new_vector_cpu[CYGPKG_HAL_SMP_CPU_MAX];
cyg_uint32 leon3_hardware_isr(CYG_ADDRWORD vector, CYG_ADDRWORD data)
{
  CYG_ADDRWORD new_vector;
  cyg_ISR_t *isr;
  cyg_uint32 isr_ret;
  cyg_uint32 pending;

  // Translate vector + pending information -> new_vector:
  //   Bank = vector
  //   Search first active bit in the pending register of that bank
  //   (Bigger = More Priority)
  //   New vector = (bank-1)*INTS_PER_BANK+source + CYGNUM_HAL_ISR_MIN
  HAL_READ_UINT32(LEON_PENDING(HAL_SMP_CPU_THIS(), vector), pending);
  cyg_uint32 i;
  for (i=31; i>0; i--) {
    if (pending & (1 << (i))) break;
  }
  new_vector = ((vector-1)*CYGNUM_HAL_INTERRUPS_PER_BANK+i) + CYGNUM_HAL_ISR_MIN;

  // Call ISR
  isr = (cyg_ISR_t*) hal_interrupt_handlers[new_vector];
  data = hal_interrupt_data[new_vector];

  isr_ret = (*isr) (new_vector, data);

  // Post DSR if necessary
  if (isr_ret & CYG_ISR_CALL_DSR) {
    (gighook[GIGHOOK_ECOS_KERNEL_INTR_POST_DSR])(
        (cyg_uint32)(hal_interrupt_objects[new_vector]),
        (cyg_uint32)(0), (cyg_uint32)(0)
    );
  }

  // Return the real ISR's return code, but without DSR posting
  return isr_ret & ~CYG_ISR_CALL_DSR;
}

#endif /* _GRSIM_ */
