//===========================================================================
//
//      hal_amba.c
//
//      SPARC Architecture LEON3-specific private variables
//
//===========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Gaisler Research (Konrad Eisele<eiselekd@web.de>)
// Contributors:
// Date:         2004
// Purpose:      private amba for LEON3 processor.
//
//####DESCRIPTIONEND####
//
//===========================================================================

#ifdef CYGPKG_KERNEL
#include <pkgconf/kernel.h>
#endif

#include <cyg/infra/cyg_type.h>         // Base types
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros
#include <cyg/infra/diag.h>

#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_smp.h>
#include <cyg/hal/hal_cache.h>

#include <cyg/hal/hal_arch.h>

#ifndef BUG
#define BUG() do { /*diag_printf("BUG() at %s %d\n", __FILE__, __LINE__);*/  while(1); } while (1)
#endif

__externC void cyg_scheduler_set_need_reschedule(void);
__externC void cyg_scheduler_timeslice_cpu(void);
__externC void cyg_scheduler_reset_timeslice_count(void);

cyg_uint32 sparc_leon3_get_cpuid(void)
{
  unsigned int id;
  __asm__ __volatile__(
    "rd     %%asr17,%0\n\t"
  : "=r" (id) : );

  return ((id >> 28) & 0xff);
}


extern __inline__ void __delay(unsigned long loops)
{
  __asm__ __volatile__("cmp %0, 0\n\t"
           "1: bne 1b\n\t"
           "subcc %0, 1, %0\n" :
           "=&r" (loops) :
           "0" (loops) :
           "cc");
}

#ifdef CYGPKG_HAL_SMP_SUPPORT

volatile unsigned int cyg_hal_smp_cpu_running = 0;
CYG_WORD32 cyg_hal_smp_cpu_running_count;
externC CYG_WORD32 *__interrupt_stack_vector[HAL_SMP_CPU_MAX];
externC CYG_WORD32 *cyg_interrupt_stack;
externC CYG_WORD32 cyg_interrupt_stack_size;

volatile int slavesem = 0;

#define CYG_HAL_CPU_MARKRUNNING(c) cyg_hal_smp_cpu_running |= (1 << c);
#define CYG_HAL_CPU_ISRUNNING(c) (cyg_hal_smp_cpu_running & (1 << c))

#define LEON_ENABLE_SET(cpu,bank)   \
   ((&LEON3_IrqCtrl_Regs_Write ->set_enable[cpu*CYGNUM_HAL_BANK_COUNT+bank]))
#define LEON_ENABLE_CLEAR(cpu,bank) \
   ((&LEON3_IrqCtrl_Regs_Write ->clear_enable[cpu*CYGNUM_HAL_BANK_COUNT+bank]))

#define get_irqmask(irq) LEON_BANK_INT(irq)

#ifndef _GRSIM_
// Get mask of a given interrupt
#define LEON_BANK_INT(x) \
          (1 << (((x) - CYGNUM_HAL_ISR_MIN) % CYGNUM_HAL_INTERRUPS_PER_BANK))

// Get bank of a given interrupt - number of banks must be power of two
#define CYGNUM_HAL_BANK_GET(_n_) \
          ((((_n_) - CYGNUM_HAL_ISR_MIN) / CYGNUM_HAL_INTERRUPS_PER_BANK) + 1)
#endif

/* master(cpu 0) coming up from arch/vector.S: after locore init, before cyg_start(). */
__externC void cyg_hal_smp_init(void) {
  volatile int slavesemold[32];
    HAL_SMP_CPU_TYPE cpu;

    do {
      HAL_TAS_SET(slavesem, slavesemold[HAL_SMP_CPU_THIS()]);
    } while (slavesemold[HAL_SMP_CPU_THIS()] == 1);

    //leon3smp_diag_printf("Master Cpu %d came up\n",HAL_SMP_CPU_THIS());
#ifndef _NO_LDD_

    unsigned long cfg = sparc_leon3_get_dcachecfg();
    //cyg_hal_smp_cpu_running = 0;

    if (cfg & ASI_LEON3_SYSCTRL_CFG_SNOOPING) {
      sparc_leon3_enable_snooping();
    }
    else {
      leon3smp_diag_printf("Caches disabled due to lack snooping\n");
      sparc_leon3_disable_cache();
    }

#else

      sparc_leon3_disable_snooping();
      sparc_leon3_disable_dcache();

#endif

    for( cpu = 0; cpu < HAL_SMP_CPU_MAX; cpu++ )  {
      /*
        if( cpu == HAL_SMP_CPU_THIS() ) {
          CYG_HAL_CPU_MARKRUNNING(cpu);
          cyg_hal_smp_cpu_running_count++;
  }
      */
  __interrupt_stack_vector[cpu] =  (CYG_WORD32 *)(((unsigned int)(cyg_interrupt_stack)) +
           (cyg_interrupt_stack_size * cpu));
        //diag_printf("Interrupt stack(%d): 0x%x-0x%x\n",cpu,
                    //__interrupt_stack_vector[cpu],
                    //(((unsigned int)__interrupt_stack_vector[cpu])+CYGNUM_HAL_COMMON_INTERRUPTS_STACK_SIZE));

    }
    CYG_HAL_CPU_MARKRUNNING(0);
    cyg_hal_smp_cpu_running_count++;

    slavesem = 0;

}

#ifdef _GRSIM_

/* release the slaves that are idling , they'll come up at cyg_hal_smp_startup() */
cyg_uint32 cyg_hal_cpu_reset(cyg_uint32 cpu, cyg_uint32 arg2, cyg_uint32 arg3)
{
  if (!CYG_HAL_CPU_ISRUNNING(cpu)) {
    //reset cpu
    LEON3_IrqCtrl_Regs->mpstatus = 1<<cpu;
  }
  return 0;
}

#else /* _GRSIM_ */

/* release the slaves that are idling , they'll come up at cyg_hal_smp_startup() */
cyg_uint32 cyg_hal_cpu_reset(cyg_uint32 cpu, cyg_uint32 arg2, cyg_uint32 arg3)
{

  if (!CYG_HAL_CPU_ISRUNNING(cpu)) {
    //reset cpu
    LEON3_IrqCtrl_Regs_Write->status = 1<<cpu;
  }
  return 0;

}

#endif /* _GRSIM_ */

cyg_uint32 cyg_hal_smp_cpu_count2idx(cyg_uint32 n) {
  int idx = -1;
  if (n >= cyg_hal_smp_cpu_running_count) {
    BUG();
  }
  do {
    idx++;
    while(! CYG_HAL_CPU_ISRUNNING(idx)) {
      idx++;
    }
  } while (n--);
  if (idx >= HAL_SMP_CPU_MAX) {
//    diag_printf("idx(%d)>=HAL_SMP_CPU_MAX(%d)\n",idx,HAL_SMP_CPU_MAX);
    BUG();
  }
  return idx;
}

cyg_uint32 cyg_hal_smp_cpu_idx2count(cyg_uint32 n) {
  int i, c = 0;
  if (n >= HAL_SMP_CPU_MAX || !CYG_HAL_CPU_ISRUNNING(n)) {
//    diag_printf("n(%d)>=HAL_SMP_CPU_MAX(%d) || !CYG_HAL_CPU_ISRUNNING(n)\n",n,HAL_SMP_CPU_MAX);
    BUG();
  }
  for (i = 0; i < n;i++) {
    if (CYG_HAL_CPU_ISRUNNING(i)) {
      c++;
    }
  }
  if (c >= cyg_hal_smp_cpu_running_count) {
//    diag_printf("c(%d)>=cyg_hal_smp_cpu_running_count(%d)\n",c,cyg_hal_smp_cpu_running_count);
    BUG(); }
  return c;
}

int cyg_hal_smp_cpu_count(void) {
  int i = 0,cpu = 0;
  for( cpu = 0; cpu < HAL_SMP_CPU_MAX; cpu++ )  {
    if (CYG_HAL_CPU_ISRUNNING(cpu)) {
      i++;
    };
  }
  return i;
}

unsigned int rmask[4] = {0, 1, 3, 7};


externC void cyg_kernel_smp_startup(void);

/* slave cpus coming from vector.S:wakeup write to mpirq->mpstatus */
__externC void cyg_hal_smp_startup(void)
{
  (gighook[GIGHOOK_ECOS_HAL_SMP_STARTUP])((cyg_uint32)(0),(cyg_uint32)(0),(cyg_uint32)(0));
}

/* slave cpus coming from vector.S:wakeup write to mpirq->mpstatus */
externC cyg_uint32 my_cyg_hal_smp_startup(cyg_uint32 arg1, cyg_uint32 arg2, cyg_uint32 arg3, ...)
{
  volatile int slavesemold[32];

   do {
     HAL_TAS_SET(slavesem, slavesemold[HAL_SMP_CPU_THIS()]);
   } while (slavesemold[HAL_SMP_CPU_THIS()] == 1);


  HAL_SMP_CPU_TYPE cpu = HAL_SMP_CPU_THIS();

/*   do { */
/*     //HAL_TAS_SET(slavesem, slavesemold);  */
/*     //} while ((slavesemold == 1) && (cyg_hal_smp_cpu_running != rmask)); */
  //do {} while (cyg_hal_smp_cpu_running != rmask[cpu]);


    CYG_HAL_CPU_MARKRUNNING(cpu);

    //leon3smp_diag_printf("Slave Cpu %d came up\n",cpu);

#ifndef _NO_LDD_

    unsigned long cfg = sparc_leon3_get_dcachecfg();
    //cyg_hal_smp_cpu_running = 0;

    if (cfg & ASI_LEON3_SYSCTRL_CFG_SNOOPING) {
      sparc_leon3_enable_snooping();
    }
    else {
      leon3smp_diag_printf("Caches disabled due to lack snooping\n");
      sparc_leon3_disable_cache();
    }

#else

      sparc_leon3_disable_snooping();
      sparc_leon3_disable_dcache();

#endif


    cyg_hal_smp_cpu_running_count++;
    slavesem = 0;
    cyg_kernel_smp_startup();

    return 0;
}


/*------------------------------------------------------------------------*/
// SMP message buffers.
// SMP CPUs pass messages to eachother via a small circular buffer
// protected by a spinlock. Each message is a single 32 bit word with
// a type code in the top 4 bits and any argument in the remaining
// 28 bits.

#define SMP_MSGBUF_SIZE 4

static struct smp_msg_t
{
    HAL_SPINLOCK_TYPE           lock;           // protecting spinlock
    volatile CYG_WORD32         msgs[SMP_MSGBUF_SIZE]; // message buffer
    volatile CYG_WORD32         head;           // head of list
    volatile CYG_WORD32         tail;           // tail of list
    volatile CYG_WORD32         reschedule;     // reschedule request
    volatile CYG_WORD32         timeslice;      // timeslice request
} smp_msg[HAL_SMP_CPU_MAX];

//extern int cpu_int_ongoing;
cyg_uint32 cyg_hal_cpu_message(cyg_uint32 cpu, cyg_uint32 msg, cyg_uint32 wait)
{
    CYG_INTERRUPT_STATE istate;
    struct smp_msg_t *m = &smp_msg[cpu];
    int i;
    //HAL_SMP_CPU_TYPE me = HAL_SMP_CPU_THIS();

    //leon3smp_diag_printf("Cpu %d send to %d: %x \n",HAL_SMP_CPU_THIS(),cpu,msg);
//
    //while(cpu_int_ongoing);
    HAL_DISABLE_INTERRUPTS( istate );

    // Get access to the message buffer for the selected CPU
    HAL_SPINLOCK_SPIN( m->lock );

    if( (msg & HAL_SMP_MESSAGE_TYPE) == HAL_SMP_MESSAGE_RESCHEDULE )
        m->reschedule++;
    else if( (msg & HAL_SMP_MESSAGE_TYPE) == HAL_SMP_MESSAGE_TIMESLICE )
        m->timeslice++;
    else
    {
        CYG_WORD32 next = (m->tail + 1) & (SMP_MSGBUF_SIZE-1);
        // If the buffer is full, wait for space to appear in it.
        // This should only need to be done very rarely.
        while( next == m->head ) {
            HAL_SPINLOCK_CLEAR( m->lock );
            for( i = 0; i < 1000; i++ );
            HAL_SPINLOCK_SPIN( m->lock );
        }
        m->msgs[m->tail] = msg;
        m->tail = next;
    }

    HAL_INTERRUPT_SMP_RAISE(cpu);

    HAL_SPINLOCK_CLEAR( m->lock );

    // If we are expected to wait for the command to complete, then
    // spin here until it does. We actually wait for the destination
    // CPU to empty its input buffer. So we might wait for messages
    // from other CPUs as well. But this is benign.

    while(wait) {
        for( i = 0; i < 1000; i++ );
        HAL_SPINLOCK_SPIN( m->lock );
        if( m->head == m->tail )
            wait = false;
        HAL_SPINLOCK_CLEAR( m->lock );
    }

    HAL_RESTORE_INTERRUPTS( istate );

  return 0;
}

int rcvd_cpu_int[2] = {0,0};
/*------------------------------------------------------------------------*/
__externC CYG_WORD32 cyg_hal_cpu_message_isr( CYG_WORD32 vector, CYG_ADDRWORD data )
{
    HAL_SMP_CPU_TYPE me = HAL_SMP_CPU_THIS();
    struct smp_msg_t *m = &smp_msg[me];
    rcvd_cpu_int[me]++;
    CYG_WORD32 ret = 1;
    CYG_INTERRUPT_STATE istate;

    HAL_DISABLE_INTERRUPTS( istate );

    HAL_SPINLOCK_SPIN( m->lock );

    HAL_REORDER_BARRIER();

    // First, mask and acknowledge the interrupt.
#ifndef _GRSIM_
    HAL_INTERRUPT_MASK( vector );
#endif //_GRSIM_

    HAL_INTERRUPT_ACKNOWLEDGE( vector );

    HAL_REORDER_BARRIER();

    if( m->reschedule || m->timeslice )
        ret |= 2;               // Ask for the DSR to be called.

    // Now pick messages out of the buffer and handle them

    while( m->head != m->tail )
    {
        CYG_WORD32 msg = m->msgs[m->head];

        switch( msg & HAL_SMP_MESSAGE_TYPE )
        {
        case HAL_SMP_MESSAGE_RESCHEDULE:
        case HAL_SMP_MESSAGE_TIMESLICE:
            ret |= 2;           // Ask for the DSR to be called.
            break;
        /*case HAL_SMP_MESSAGE_MASK:
            break;
        case HAL_SMP_MESSAGE_UNMASK:
            break;
        case HAL_SMP_MESSAGE_REVECTOR:
            break;*/
        default:
          (gighook[GIGHOOK_CPU_MSG])((cyg_uint32)(msg),(cyg_uint32)(0),(cyg_uint32)(0));
        }

        // Update the head pointer after handling the message, so that
        // the wait in cyg_hal_cpu_message() completes after the action
        // requested.
        m->head = (m->head + 1) & (SMP_MSGBUF_SIZE-1);
    }

    HAL_REORDER_BARRIER();

#ifndef _GRSIM_
    if (ret == 1)
    {
      /* Unmask interrupt before leaving. */
      HAL_INTERRUPT_UNMASK( vector );
    }
#endif //_GRSIM_

    HAL_REORDER_BARRIER();

    HAL_SPINLOCK_CLEAR( m->lock );
    /*leon3_smpirq_lock();
    cpu_int_ongoing &= ~(1<<me);
    leon3_smpirq_unlock();
*/
    HAL_RESTORE_INTERRUPTS( istate );


    return ret;

}


/*------------------------------------------------------------------------*/
// CPU message DSR.
// This is only executed if the message was
// HAL_SMP_MESSAGE_RESCHEDULE. It calls up into the kernel to effect a
// reschedule.

__externC CYG_WORD32 cyg_hal_cpu_message_dsr( CYG_WORD32 vector, CYG_ADDRWORD data )
{
    HAL_SMP_CPU_TYPE me = HAL_SMP_CPU_THIS();
    struct smp_msg_t *m = &smp_msg[me];
    CYG_INTERRUPT_STATE istate;
    CYG_WORD32 reschedule, timeslice;

    HAL_DISABLE_INTERRUPTS( istate );
    HAL_SPINLOCK_SPIN( m->lock );

    reschedule = m->reschedule;
    timeslice = m->timeslice;
    m->reschedule = m->timeslice = 0;

#ifndef _GRSIM_
    /* Unmask interrupt before leaving. */
    HAL_INTERRUPT_UNMASK( vector );
#endif //_GRSIM_

    HAL_SPINLOCK_CLEAR( m->lock );
    HAL_RESTORE_INTERRUPTS( istate );

    //while( reschedule-- )
    if ( reschedule > 0 )
    {
      //leon3smp_diag_printf("Cpu %d received reshedule\n",HAL_SMP_CPU_THIS());
      cyg_scheduler_set_need_reschedule();
    }
#ifdef CYGSEM_KERNEL_SCHED_TIMESLICE
    if ( timeslice > 0 )
    //while( timeslice-- )
    {
      //leon3smp_diag_printf("Cpu %d received timeslice\n",HAL_SMP_CPU_THIS());
      cyg_scheduler_timeslice_cpu();
    }
#endif

    return 0;
}



#ifdef _NO_LDD_
/*------------------------------------------------------------------------
 ** \brief  hal_tas_set  workaround for lewis
 */
externC CYG_WORD32 __hal_tas_set(volatile CYG_WORD32 * _tas_, CYG_WORD32  __old)
{
  /* 5 * nop  + swapa + nop -> 7 words of a 8 word cacheline */
  __asm__ volatile ("nop;nop;nop;nop;nop;");

  __asm__ volatile ("swapa [%2] %3, %0" :
            "=r" (__old) :
            "0" (__old),
            "r" (_tas_),
            "i" (1));

  __asm__ volatile ("nop;");

  return (__old);

}
#endif // _NO_LDD_

#endif //CYGPKG_HAL_SMP_SUPPORT
