/**
 * \addtogroup gighook
 */
/*@{*/

/***************************************************
 * Header name: gighook_api.h
 *
 * Copyright 2006 Gigle Semiconductor as an unpublished work.
 * All Rights Reserved.
 *
 *  The information contained herein is confidential
 * property of Company. The user, copying, transfer or
 * disclosure of such information is prohibited except
 * by express written agreement with Company.
 *
 * First written on 02/10/2007 by I- Barrera
 *
 ***************************************************/
/** \file gighook_api.h
 *
 * \brief This is the exported API of GigHook
 *
 *
 **************************************************/

/***************************************************
 *                 Public Constants Section
 ***************************************************/

/***************************************************
 *                 Public Typedefs Section
 ***************************************************/
#ifndef GIGHOOK_H_
#define GIGHOOK_H_

/** \brief Hook number. */
typedef enum
{
 GIGHOOK_ECOS_INSTRUMENT = 0,            /**< Hook for eCos instrumentation. */
 GIGHOOK_ECOS_CYG_START,                 /**< Hook for the cyg_start function. */
 GIGHOOK_ECOS_MAIN,                      /**< Hook for the main function. */
 GIGHOOK_ECOS_AMBA_INIT_POST,            /**< Hook for the amba_init function. */
 GIGHOOK_ECOS_HAL_START,                 /**< Hook for the cyg_hal_start function. */
 GIGHOOK_ECOS_HAL_SMP_STARTUP,           /**< Hook for the cyg_hal_smp_startup function. */
 GIGHOOK_ECOS_KERNEL_SCHED_START,        /**< Hook for the cyg_scheduler_start function. */
 GIGHOOK_ECOS_KERNEL_SCHED_START_ALL,    /**< Hook for the cyg_scheduler_start_all function. */

 GIGHOOK_ECOS_HAL_INTERRUPT_SMP_INIT,    /**< Hook for the HAL_INTERRUPT_SMP_INIT function. */
 GIGHOOK_ECOS_HAL_INTERRUPT_MASK,        /**< Hook for the HAL_INTERRUPT_MASK function. */
 GIGHOOK_ECOS_HAL_INTERRUPT_UNMASK,      /**< Hook for the HAL_INTERRUPT_UNMASK function. */
 GIGHOOK_ECOS_HAL_INTERRUPT_SET_CPU,     /**< Hook for the HAL_INTERRUPT_SET_CPU function. */
 GIGHOOK_ECOS_HAL_INTERRUPT_GET_CPU,     /**< Hook for the INTERRUPT_GET_CPU function. */
 GIGHOOK_ECOS_HAL_INTERRUPT_ACKNOWLEDGE, /**< Hook for the HAL_INTERRUPT_ACKNOWLEDGE function. */
 GIGHOOK_ECOS_HAL_INTERRUPT_SMP_RAISE,   /**< Hook for the HAL_INTERRUPT_SMP_RAISE function. */
 GIGHOOK_ECOS_KERNEL_INTR_POST_DSR,      /**< Hook for the post_dsr function. */

 GIGHOOK_ECOS_KERNEL_CREATE_CPU_INT,     /**< Hook for creating the CPU message interrupt. */
 GIGHOOK_ECOS_HAL_SMP_CPU_MESSAGE,       /**< Hook for the HAL_SMP_CPU_*_INTERRUPT functions. */
 GIGHOOK_ECOS_HAL_SMP_CPU_START,         /**< Hook for the HAL_SMP_CPU_START function. */

 GIGHOOK_GIGOS_TIME,                     /**< Hook for GigOS Time API */
 GIGHOOK_GIGOS_MUTEX,                    /**< Hook for GigOS Mutex API */
 GIGHOOK_GIGOS_SEMAPHORE,                /**< Hook for GigOS Semaphore API */
 GIGHOOK_GIGOS_MQUEUE,                   /**< Hook for GigOS MQueue API */

 GIGHOOK_GIGLOG_ASSERT,                  /**< Hook for the GigLog Assertion API. */
 GIGHOOK_GIGLOG_PRINTF,                  /**< Hook for the GigLog Log API. */

 GIGHOOK_CONSOLE_WRITE,                  /**< Hook for the Console Write. */
 GIGHOOK_CONSOLE_READ,                   /**< Hook for the Console Read. */
 GIGHOOK_CONSOLE_HOUSEKEEP,              /**< Hook for the Console Housekeeping. */

 GIGHOOK_HAL_WDOG_GETNV,                 /**< Hook for the HalWatchdogGetNV function. */
 GIGHOOK_HAL_WDOG_ERASENV,               /**< Hook for the HalWatchdogEraseNV function. */
 GIGHOOK_HAL_WDOG_STORENV,               /**< Hook for the HalWatchdogStoreNV function. */

 GIGHOOK_PARAM_CONFIG_GET_VALUE,         /**< Hook for the ParamConfigGetValue function. */

 GIGHOOK_DEBUG_COUNTER_INC,              /**< Hook for the DebugCounterInc function. */
 GIGHOOK_FILESYSTEM_OPEN,                /**< Hook for the FilesystemOpen function */
 GIGHOOK_FILESYSTEM_READWRITE,           /**< Hook for the FilesystemRead/Write functions */
 GIGHOOK_FILESYSTEM_CLOSE,               /**< Hook for the FilesystemClose function */
 GIGHOOK_CPU_MSG,                         /**< Hook for unknwon CPU message received */

 GIGHOOK_N_HOOKS                         /**< Number of hooks. */
}
t_gighook_hook;

/***************************************************
 *                 Public Tables Section
 ***************************************************/

/** \brief Hook table. */
extern cyg_uint32 (*gighook[GIGHOOK_N_HOOKS+1])(cyg_uint32 arg1, cyg_uint32 arg2, cyg_uint32 arg3, ...);

/***************************************************
 *                 Public Function Prototypes Section
 ***************************************************/

#endif

/*@}*/
