/**
 * \addtogroup common
 */
/*@{*/

/***************************************************
 * Header name: gigtypes.h
 *
 ***************************************************/
/** \file gigtypes.h
 *
 * \brief This contains the definition of types
 *
 * Only general types that will be used for several
 * modules must be defined here
 *
 **************************************************/

#ifndef GIGTYPES_H_
#define GIGTYPES_H_

/***************************************************
 *                 Public Defines Section
 ***************************************************/

/** \brief Boolean type */
typedef enum {
  GIG_FALSE = 0,    /**< Boolean false */
  GIG_TRUE = 1      /**< Boolean true */
} t_bool;

/** \brief Number of possible Boolean values */
#define GIG_N_SWITCH   (2)

/** \brief Signed 8-bit integer */
typedef   signed char  t_int8;

/** \brief Unsigned 8-bit integer */
typedef unsigned char  t_uint8;

/** \brief Signed 16-bit integer */
typedef   signed short t_int16;

/** \brief Unsigned 16-bit integer */
typedef unsigned short t_uint16;

/** \brief Signed 32-bit integer */
typedef   signed int   t_int32;

/** \brief Unsigned 32-bit integer */
typedef unsigned int   t_uint32;

/** \brief Signed 64-bit integer */
typedef   signed long long  t_int64;

/** \brief Unsigned 64-bit integer */
typedef unsigned long long  t_uint64;

/** \brief type for names */
typedef char* t_name ;

/** \brief type for integer/pointer data */
typedef long t_intptr ;

/** \brief type for unsigned integer/pointer data */
typedef unsigned long t_uintptr ;

/** \brief Swap two bytes inside a half word */
#define swap16(x) (              \
  (t_uint16)(((t_uint16)(x) & 0xff) << 8 | ((t_uint16)(x) & 0xff00) >> 8)  \
)

/** \brief Swap four bytes inside a word */
#define swap32(x) (              \
    (t_uint32)(((t_uint32)(x) & 0xff) << 24 |        \
    ((t_uint32)(x) & 0xff00) << 8 | ((t_uint32)(x) & 0xff0000) >> 8 |  \
    ((t_uint32)(x) & 0xff000000) >> 24)    \
)

#ifndef _UNIT_TEST_

/** \brief Implementation of htons() function.
 *
 * Straightforward, as Leon3 is a big-endian CPU. */
#define htons(x) (x)

/** \brief Implementation of ntohs() function.
 *
 * Straightforward, as Leon3 is a big-endian CPU. */
#define ntohs(x) (x)

/** \brief Implementation of htonl() function.
 *
 * Straightforward, as Leon3 is a big-endian CPU. */
#define htonl(x) (x)

/** \brief Implementation of ntohl() function.
 *
 * Straightforward, as Leon3 is a big-endian CPU. */
#define ntohl(x) (x)

#else

/** \brief Implementation of htons() function. */
#define htons(x) swap16(x)

/** \brief Implementation of ntohs() function. */
#define ntohs(x) swap16(x)

/** \brief Implementation of htonl() function. */
#define htonl(x) swap32(x)

/** \brief Implementation of ntohl() function. */
#define ntohl(x) swap32(x)

#endif /* _UNIT_TEST_ */

#else /*GIGTYPES_H_*/
#error "Header file __GIGTYPE_H__ has already been included!"
#endif /*GIGTYPES_H_*/

/*@}*/

