/////////////////////////////////////////////////////////////////////
////                                                             ////
////  WISHBONE Memory and Bus Controller global definitions      ////
////                                                             ////
////                                                             ////
////  Author: Hakan Fouren                                       ////
////          hakan.fouren@gigle.biz                             ////
////                                                             ////
/////////////////////////////////////////////////////////////////////
////                                                             ////
//// Copyright (C) 2005 Gigle Semiconductor                      ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

//  SVN tag information
//
//  $Id$
//

`define	BYTE_WIDTH		8

`define WB_ADDR_WIDTH		32
`define WB_ADDR_OFFSET		2
`define	WB_DATA_WIDTH		32
`define	WB_ADDR_RANGE		`WB_ADDR_WIDTH-1:`WB_ADDR_OFFSET
`define WB_SEL_WIDTH		(`WB_DATA_WIDTH/`BYTE_WIDTH)

`define RAM_SELECT		31
`define RAM_VALUE		1'b0
`define RAM_ADDR_WIDTH		20
`define RAM_ADDR_OFFSET		4
`define RAM_DATA_WIDTH		128
`define RAM_SEL_WIDTH		(`RAM_DATA_WIDTH/`BYTE_WIDTH)
`define RAM_HIT_MASK		({{`RAM_ADDR_WIDTH{1'b1}}, \
                                  {`RAM_ADDR_OFFSET{1'b0}}} | \
                                 {1'b1,{`RAM_SELECT{1'b0}}})
`define	RAM_ADDR_RANGE		`RAM_ADDR_WIDTH-1:`RAM_ADDR_OFFSET
`define RAM_WORD_SELECT		`RAM_ADDR_OFFSET-1:`WB_ADDR_OFFSET

`define SLAVE_SELECT		30:27

`define SPI_SELECT_ADDR		32'h80000018
`define SPI_CONTROL_ADDR	32'h80000010
`define SPI_DATA_ADDR		32'h80000000

`define FLASH_SELECT_VALUE	32'h00000001
`define FLASH_DESELECT_VALUE	32'h00000000
`define FLASH_ADDRESS_WIDTH	32
`define FLASH_ADDRESS_MASK	32'h00FFFFF0
`define FLASH_READ_CMD		32'h03000000
`define FLASH_ADDRESS_CMD	{20'h0,spi_lsb_i,spi_txedge_i,spi_rxedge_i,\
                                 9'd`FLASH_ADDRESS_WIDTH}
`define FLASH_DATA_CMD		{20'h0,spi_lsb_i,spi_txedge_i,spi_rxedge_i,\
				 9'd`RAM_DATA_WIDTH}
`define FLASH_GO_MASK		32'h00000100

`define WB_SLAVES		16
`define WB_SLAVE_SELECT_RANGE	30:27 // # bits should cover WB_SLAVES

// Boot code is the first thing read from the instruction masters
// Assembly code is:
// l.movhi r1,     0x8000	; Start high address (CPU2: 0x0000)
// l.ori   r1, r1, 0x0000	; Start low address  (CPU2: 0x0000)
// l.jr    r1			; Jump to Flash
// l.nop           0x0000	; Fill delay slot
// Boot code assembly:               l.nop    l.jr     l.ori    l.movhi 
`define CPU1_BOOT_CODE		128'h15000000_44000800_a8210000_18208000
`define CPU2_BOOT_CODE		128'h15000000_44000800_a8210000_18200000
`define CPU_BOOT_ADDRESS	32'h00000100

// Defines that shouldn't change

`define WB_CTI_BITS		3
`define WB_CTI_SINGLE		3'b000
`define WB_CTI_CONSTANT		3'b001
`define WB_CTI_SEQUENTIAL	3'b010
`define WB_CTI_END		3'b111

`define WB_BTE_BITS		2
`define WB_BTE_LINEAR		2'b00
`define WB_BTE_4WRAP		2'b01
`define WB_BTE_8WRAP		2'b10
`define WB_BTE_16WRAP		2'b11
